#!/usr/bin/env python3

class Token():
    def __init__(self, content: str):
        self.content = content
    def __repr__(self):
        content = self.content \
            .replace('\\', '\\\\') \
            .replace('"', '\"') \
            .replace('\t', '\\t') \
            .replace('\n', '\\n')
        return f'{type(self).__name__}<{content}>'
class PlainToken(Token):
    pass
class MetaToken(Token):
    pass

def tokenize(template: str) -> list[Token]:
    """
    Parse a template into a list of tokens. Tokens in the resulting list
    are either of type PlainToken or of type MetaToken. PlainToken is used
    for literal content. MetaToken is used for tokens that have a special
    meaning according to the template syntax, such as $var$, $if(var)$,
    $for(var)$, $endfor$ or $endif$.
    """
    def split_at(string: str, sep: str) -> tuple[str,str]:
        if sep not in string:
            return string, ""
        return string.split(sep, 1)
    current = list()
    rest = template
    while len(rest) > 0:
        head, rest = split_at(rest, '$')
        current.append(head)
        if rest[:1] == '$':             # Literal $
            current.append("$");
            rest = rest[1:]
        elif rest[:2] == '--':          # Line comment
            _head, rest = split_at(rest, '\n')
        elif rest[:1] == '{':
            yield PlainToken("".join(current)); current = list()
            head, rest = split_at(rest[1:], '}')
            yield MetaToken(head)
        else:                           # Dollar-enclosed MetaToken
            yield PlainToken("".join(current)); current = list()
            head, rest = split_at(rest, '$')
            yield MetaToken(head)
    if current: yield PlainToken("".join(current))



#!/usr/bin/env python3

import re
from .tokenizer import *
from .syntax import *

# Note: Internal __parse* functions consume (part of) the tokens they
# receive as their argument.
def __parse_if(tokens: list[Token]) -> If:
    l = tokens
    assert l[0].content[:3] == 'if(' and l[0].content[-1:] == ')'
    condition = l.pop(0).content[3:-1]
    thentree, end = __parse(l, endtokens=[r'elseif\(.*\)', 'else', 'endif'])
    if end == 'else':
        elsetree, end = __parse(l, endtokens=['endif'])
    elif end.startswith('elseif('):
        assert end[:4] == 'else'
        l.insert(0, MetaToken(end[4:]))
        elsetree, _ = __parse(l, endtokens=['endif'])
    else:
        elsetree = Str('')
    return If(condition, thentree, elsetree)
def __parse_for(tokens: list[Token]) -> For:
    l = tokens
    assert l[0].content[:4] == "for(" and l[0].content[-1:] == ")"
    it = l.pop(0).content[4:-1]
    main, end = __parse(l, endtokens=['sep', 'endfor'])
    sep = __parse(l, endtokens=['endfor'])[0] if end == 'sep' else None
    return For(it, main, sep)

def __parse(tokens: list[Token], endtokens: list[str]=None) -> tuple[Tree,str]:
    """Returns a tuple (tree, matching_endtoken)"""
    l = tokens
    res = list()
    while len(l) > 0:
        if type(l[0]) == PlainToken:
            res.append(Str(l.pop(0).content))
        elif type(l[0]) == MetaToken:
            if endtokens != None:
                for t in endtokens:
                    if re.fullmatch(t, l[0].content):
                        return (Tree(res), l.pop(0).content)
            if re.fullmatch(r'if\(.*\)', l[0].content):
                res.append(__parse_if(l))
            elif re.fullmatch(r'for\(.*\)', l[0].content):
                res.append(__parse_for(l))
            else:
                res.append(Var(l.pop(0).content))
    return (Tree(res), None)
def parse_tokens(tokens: list[Token]) -> Tree:
    """
    Parse a list of Tokens into an actual syntax Tree.
    """
    tokens = list(filter(
        lambda x: not (type(x) == PlainToken and x.content == ""),
        tokens
    ))
    tree, _ = __parse(tokens)
    return tree



#!/usr/bin/env python3

from typing import Union as _Union
from copy import deepcopy

def _getvar(metadata: dict, varname: str) -> _Union[str,list,dict]:
    d = metadata
    for key in varname.split('.'):
        try:
            d = d.__getitem__(key)
        except:
            return ''
    return d
def _setvar(metadata: dict, newvars: dict) -> dict:
    newdict = deepcopy(metadata)
    for varname, val in newvars.items():
        d = newdict
        for key in varname.split('.')[:-1]:
            if type(d.get(key, '')) != dict:
                d[key] = dict()
            d = d[key]
        d[varname.split('.')[-1]] = val
    return newdict

class TreeElement():
    def __repr__(self) -> str:
        res = list()
        res.append(f'{type(self).__name__}<')
        res.extend(['    ' + l for l in repr(self.__dict__).splitlines()])
        res.append('>')
        return '\n'.join(res)
class Str(TreeElement):
    def __init__(self, content: str):
        self.content = content
    def apply(self, metadata: dict) -> str:
        return self.content
class Var(TreeElement):
    def __init__(self, name: str):
        self.name = name
    def apply(self, metadata: dict) -> str:
        val = _getvar(metadata, self.name)
        if type(val) == list:   return ''.join(map(str, val))
        elif type(val) == dict: return 'true'
        else:                   return str(val)
class If(TreeElement):
    def __init__(self, condition: str, thenarm: TreeElement, elsearm: TreeElement):
        self.condition = condition
        self.thenarm = thenarm
        self.elsearm = elsearm
    def apply(self, metadata: dict) -> None:
        if _getvar(metadata, self.condition) == '':
            return self.elsearm.apply(metadata)
        else:
            return self.thenarm.apply(metadata)
class For(TreeElement):
    def __init__(self, it: str, main: TreeElement, sep: TreeElement):
        self.it = it
        self.main = main
        self.sep = sep
    def apply(self, metadata: dict) -> str:
        res = list()
        data = _getvar(metadata, self.it)
        if type(data) != list:
            res.append(self.main.apply(_setvar(metadata, {'it': data})))
        else:   # type(data) == list
            for i, it in enumerate(data):
                res.append(self.main.apply(_setvar(
                    metadata, {'it': it, self.it: it}
                )))
                if self.sep != None and i < len(data)-1:
                    res.append(self.sep.apply(_setvar(
                            metadata, {'it': it, self.it: it}
                    )))
        return "".join(res)
class Tree(TreeElement):
    def __init__(self, tokens: list[TreeElement]):
        self.tokens = tokens
    def apply(self, metadata: dict) -> str:
        return "".join([x.apply(metadata) for x in self.tokens])



#!/usr/bin/env python3

from .tokenizer import tokenize
from .parser import parse_tokens
from .syntax import Tree as _Tree

class Template():
    def __init__(self, template: str):
        self.tree: _Tree = parse_tokens(tokenize(template))
    def apply(self, metadata: dict) -> str:
        return self.tree.apply(metadata)
    def __repr__(self) -> str:
        return f'Template:{self.tree}'
def parse_template(template: str) -> Template:
    return Template(template)
def apply_template(template: str, metadata: dict) -> str:
    return parse_template(template).apply(metadata)



from setuptools import setup, find_packages

with open('readme.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='dollar_templates',
    version='1.0.1',
    packages=['dollar_templates'],
    python_requires='>=3.9',
    author='Frank Seifferth',
    author_email='frankseifferth@posteo.net',
    description='Pandoc style dollar templates for python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/seifferth/dollar_templates',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License '
                                   'v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
)
